/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2019-2019. All rights reserved.
 * Description:
 *      This file is the implement of class BagFile.
 *      BagFile will create a bag file and provide read or write operation
 * Create: 2019-11-30
 * Notes: NA
 */
#ifndef RTF_BAG_FILE_H
#define RTF_BAG_FILE_H

#include <iostream>
#include <mutex>
#include <set>
#include <sys/types.h>
#include <tuple>

#include "ara/core/map.h"
#include "ara/core/string.h"
#include "ara/core/vector.h"
#include "rtf/internal/compressor.h"
#include "rtf/internal/RtfBagStructs.h"
#include "rtf/internal/RtfBuffer.h"
#include "rtf/internal/RtfChunkManager.h"
#include "rtf/internal/RtfHeader.h"
#include "rtf/maintaind/impl_type_eventinfo.h"
#include "rtf/rtfbag/RtfBagInfo.h"

namespace rtf {
namespace rtftools{
namespace utils{
class NetProcess;
}
}
namespace rtfbag {
extern const uint32_t CURRENT_BAG_VERSION;
extern const std::string CURRENT_BAG_VERSION_STR;
enum class FileCompressionType : uint32_t {
    NONE     = 0U,    // 未压缩
    GZIP     = 1U,    // GZIP
    ZLIB     = 2U,    // ZLIB
};

enum class FileErrorCode : int32_t {
    SUCCESS     = 0,             // 成功
    WRITE_FAIL  = -1,            // 写失败
    READ_FAIL   = -2,            // 读失败
    COMPRESS_ERROR   = -3,       // 压缩失败
    DECOMPRESS_ERROR = -4,       // 解压失败
};

class PlayedEventIndex {
    public:
        PlayedEventIndex(rtf::stdtype::uint16_t serviceId, rtf::stdtype::uint16_t instanceId) noexcept
            : serviceId_(serviceId), instanceId_(instanceId) {}
        PlayedEventIndex() = default;
        friend bool operator<(const PlayedEventIndex& lhs, const PlayedEventIndex& rhs) noexcept;
        friend bool operator==(const PlayedEventIndex& lhs, const PlayedEventIndex& rhs) noexcept;
        rtf::stdtype::uint16_t GetServiceId() const noexcept;
        rtf::stdtype::uint16_t GetInstanceId() const noexcept;
        rtf::stdtype::uint16_t GetEventId() const noexcept;
        rtf::stdtype::uint16_t GetDomainId() const noexcept;
        rtf::stdtype::String GetTopicName() const noexcept;
        rtf::stdtype::String GetNetworkSegment() const noexcept;
        rtf::maintaind::DriverType GetDriverType() const noexcept;
        void SetDDSEventIndex(rtf::maintaind::DriverType driver,
                              const rtf::stdtype::String& topicName, const rtf::stdtype::String& networkSegment,
                              rtf::stdtype::uint16_t domainId) noexcept;
        void SetSOMEIPEventIndex(rtf::maintaind::DriverType driver, const rtf::stdtype::String& networkSegment, rtf::stdtype::uint16_t eventId) noexcept;
    private:
        rtf::maintaind::DriverType driver_{rtf::maintaind::DriverType::DDS};
        rtf::stdtype::uint16_t serviceId_{0U};
        rtf::stdtype::uint16_t instanceId_{0U};
        rtf::stdtype::uint16_t eventId_{0U};
        rtf::stdtype::String topicName_{""};
        rtf::stdtype::String networkSegment_;
        rtf::stdtype::uint16_t domainId_{1000U};
};

class RtfBagFile {
    friend class RtfView;
public:
    RtfBagFile() noexcept;
    ~RtfBagFile();
    RtfBagFile(RtfBagFile const&) = delete;
    RtfBagFile(RtfBagFile&&) = delete;
    RtfBagFile& operator=(RtfBagFile const&) = delete;
    RtfBagFile& operator=(RtfBagFile&&) = delete;
    void CopyOtherBagFileReadResult(const RtfBagFile& other);
    /*
     * @Description: 打开文件并写入文件头相关信息
     * @param: input ara::core::String const& fileName
     * @return: 0: 成功 else: 失败
     */
    int32_t OpenWrite(ara::core::String const& fileName, ara::core::String const& filePath, 
                      ara::core::String const& file);

    /*
     * @Description: 关闭文件写入，此时会写入ConnetionInfo，ChunkInfo等信息
     * @param: none
     * @return: 0: 成功 else: 失败
     */
    FileErrorCode CloseWrite();

    FileErrorCode Write(rtf::rtfbag::EventMsg const& eventMsg, uint64_t const time, 
                        uint8_t const *msgBuff, uint32_t buffLen);

    int32_t OpenRead(ara::core::String const& fileName);
    int32_t OpenReadActiveFile(ara::core::String const& fileName);
    bool CloseRead();
    FileErrorCode ReadMsgDataIntoStream(MessageIndex const& msgIndex, RtfBuffer& buffer, const std::uint32_t len = 0U);
    bool ReadMsgDataSize(MessageIndex const& msgIndex, uint32_t& size) const;
    /*
     * @Description: 获取当前bag文件的文件大小
     * @param: none
     * @return: uint64_t 文件大小
     */
    uint64_t GetFileSize() const noexcept;

    /*
     * @Description: 获取当前打开文件的文件名称
     * @param: none
     * @return: ara::core::String 文件名称
     */
    ara::core::String GetFileName() const;
    void SetMaxChunkSize(uint32_t chunkSize) noexcept;

    /*
     * @Description: 获取bag文件头信息
     * @param:
     * @return: true: 成功 false： 失败
     */
    bool GetFileHeaderInfo(BagFileHeadInfo& headerInfo);

    void SetStartTime();
    void SetStopTime();
    bool CheckIsSameBag(const RtfBagFile &other) noexcept;
    bool OpenToReadBagFile(ara::core::String const& fileName);
    void SetVersion(const uint32_t version) noexcept;
    uint32_t GetVersion() const noexcept;
    FILE* file_;
    void SetCompressionType(const FileCompressionType compressionType) noexcept;
    FileCompressionType GetCompressionType() const noexcept;
private:
    int32_t InitCompressor(const bool isWrite);
    bool CloseFile();
    // write chunk
    bool StartChunkWriting(uint64_t time);
    FileErrorCode StopChunkWriting();
    void ClearChunkInfo() noexcept;
    void ClearCurChunkNewEvent() noexcept;
    FileErrorCode TransferChunkToStream();

    void ReleaseBuf(uint8_t* buf, const uint32_t buflength, const ara::core::String& str) const noexcept;
    bool WriteChunk();
    bool WriteChunkForMsgs(const OutChunkHeader& chunkHeader,
                           std::queue<OutMessage>& queueBuff) const;
    bool WriteChunkForMsgsForCompress(const OutChunkHeader& chunkHeader);
    bool WriteChunkForChunkIndex(std::queue<OutMessage>& queueChunkIndex) const;
    void ResetBagFile() noexcept;
    // write record
    bool WriteVersion() const;
    bool WriteFileHeaderRecord();
    void SetFileHeaderFieldName(ara::core::Vector<ara::core::String>& fieldNameString) const;
    bool WriteRecordHeader(ara::core::Vector<ara::core::String> const& fieldName,
        ara::core::Vector<ara::core::String> const& fieldValue) const;
    bool WriteChunkInfoRecords() const;
    bool WriteChunkInfoRecord(const ChunkInfo& chunkInfoRecord) const;
    bool WriteTempChunkInfoRecord(const ChunkInfo& chunkInfoRecord);
    bool WriteConnectionRecords() const;
    bool WriteConnectionRecord(const Connection& connRecord) const;
    bool WriteEventInfo(ara::core::Vector<ara::core::String>& fieldValueString,
                        const Connection& connRecord) const;
    bool WriteSerializationConfig(ara::core::Vector<ara::core::String>& fieldValueString,
                                  const Connection& connRecord) const;
    bool WriteTempConnectionRecord(const Connection& connRecord);
    bool WriteChunkHeaderRecord(uint32_t currChunkSize, uint32_t currChunkCompressedSize);
    bool WriteMessageRecord(uint32_t connnectionId,
        uint64_t time, uint8_t const *msgBuff, uint32_t msgLen);
    bool WriteChunkIndexRecords();
    bool WriteChunkIndexRecord(uint32_t id, const std::multiset<MessageIndex> &indexSet);
    bool TransferMessageIndexToBuff(uint8_t* pBuff, const uint32_t dataLen,
                                    const std::multiset<MessageIndex> &indexSet) const;
    bool WriteChunkEndInfoRecord();
    uint32_t GetRecordHeaderLen(ara::core::Vector<ara::core::String> const& fieldName,
        ara::core::Vector<ara::core::String> const& fieldValue) const noexcept;
    uint32_t GetConnectionId(rtf::rtfbag::EventMsg const& eventMsg);

    // read record
    std::tuple<bool, ino_t, time_t> GetBagChangeInfo(const ara::core::String &realPath) const noexcept;
    bool StartReading();
    bool StartReadChunkIndexRecord();
    bool ReadVersion();
    bool ReadFileHeaderRecord();
    bool ReadConnectionRecord();
    bool ReadConnectionFeild(const ReadMap &readMap, Connection &connection);
    bool ReadConnectionFeildAppend(const ReadMap &readMap, Connection &connection) const;
    bool ReadConnectionEventInfo(const ReadMap &readMap, Connection &connection) const;
    bool ReadConnectionSerializeConfig(const ReadMap &readMap, Connection &connection) const;

    bool ReadChunkInfoRecord();
    bool ReadChunkIndexRecord();
    bool ReadChunkHeader(ChunkHeader& chunkHeader) const;
    bool OptionMatch(const ReadMap& readMap, const uint8_t opt) const;
    // read header
    template<typename T>
    bool ReadField(const ReadMap& readMap, const ara::core::String& field, T* data) const;
    bool ReadField(const ReadMap& readMap, const ara::core::String& field, ara::core::String& data) const;
    bool ReadHeader(RtfHeader& header) const;
    bool ReadHeaderFromStream(RtfHeader& header, const uint8_t* buff) const;
    FileErrorCode ReadMsgDataFromBuffer(RtfBuffer& buffer, const uint64_t offset, const std::uint32_t len) const;
    FileErrorCode ReadMsgDataFromStream(RtfBuffer& buffer, MessageIndex const& msgIndex, const std::uint32_t len);
    FileErrorCode DecompressChunkToStream(MessageIndex const& msgIndex);
    void UpdateVersion(BagFileHeadInfo& headerInfo) const;

    bool CheckSystemIsAos() const;
    void GetFileHeaderRecordTimeInfo(BagFileHeadInfo& headerInfo) const noexcept;
    bool StartReadingActiveFile();
    bool ReadConnectionRecordFromActive();
    bool ReadChunkInfoRecordFromActive();
    bool ReadChunkEndInfo(const ChunkHeader& chunkHeader);
    void SetBagFileMessageInfo(BagFileMessageInfo &msgInfo, const Connection& conn);
    bool GetInfoFromOldBag(BagFileHeadInfo& headerInfo);
    bool GetInfoFromOptimizedBag(BagFileHeadInfo& headerInfo);
    void SetEventSumInfo();
    void SetChunkConSumInfo(rtf::rtfbag::EventMsg const& eventMsg, uint64_t const time);
    PlayedEventIndex GetEventIndex(rtf::rtfbag::EventMsg const& eventMsg) const noexcept;
    ara::core::String GetNetSegment(const ara::core::String& ip, 
                                    const ara::core::String& maskValue) const noexcept;
    bool WriteConnectionRecordFieldName(ara::core::Vector<ara::core::String>& fieldNameStr) const;
    bool WriteConnectionRecordFieldValue(ara::core::Vector<ara::core::String>& fieldValueStr,
                                         const Connection& connection) const;
    bool OpenReadForBagHeaderInfo(ara::core::String const& fileName);

    ara::core::String fileName_;
    uint32_t maxChunkSize_;
    uint64_t fileSize_;
    uint32_t chunkCount_;
    uint32_t connCount_;
    uint64_t connPos_;
    uint64_t nextChunkPos_;
    uint64_t chunkInfoRecordPos_;
    uint32_t chunkStatus_;
    BagFileHeader bagFileHeader_;
    uint64_t bagFileHeaderPos_;
    uint32_t version_;
    uint32_t writeVersion_;
    FileCompressionType compressionType_;
    FileCompressionType writeCompressionType_;
    uint32_t compressedSize_;
    ara::core::String bagVersion_;  // current bag version
    bool fileWriteOpen_;
    bool fileReadOpen_;

    // Current chunk
    bool chunkOpen_;
    uint64_t currChunkDataStartPos_;
    uint64_t currChunkDataEndPos_;
    ChunkInfo currChunkInfo_;
    ara::core::Vector<ChunkInfo> chunks_;

    uint64_t currentReadChunkDataPos_;
    uint8_t* curRdChunkBuff_;
    ara::core::Map<PlayedEventIndex, uint32_t> eventConnectionId_;

    std::set<PlayedEventIndex> eventRecorded_;
    ara::core::Map<PlayedEventIndex, Connection> eventConnection_;
    ara::core::Map<PlayedEventIndex, ChunkConSumInfo> eventConInChunk_;
    ara::core::Map<uint32_t, Connection> connections_;
    uint32_t currentChunkNewEventCount_;
    uint32_t currentChunkConnectionCount_;

    ara::core::Map<uint32_t, std::multiset<MessageIndex>> connectionIdIndex_;
    ara::core::Map<uint32_t, std::multiset<MessageIndex>> currChunkIndex_;

    uint64_t startRecordRealTime_;
    uint64_t startRecordVirtualTime_;
    uint64_t stopRecordRealTime_;
    uint64_t stopRecordVirtualTime_;
    mutable std::mutex multiRead_;
    std::tuple<bool, ino_t, time_t> bagChangInfo_;
    std::shared_ptr<rtf::RtfChunkManager> RtfChunkManager_;
    const uint32_t bagVer110 = 110U;
    const uint32_t bagVer120 = 120U;
    const uint32_t bagVer130 = 130U;
    const uint32_t bagVer140 = 140U;
    const uint32_t bagVer150 = 150U;
    const uint32_t bagVer160 = 160U;
    const uint32_t bagVer170 = 170U;
    const uint32_t bagVer180 = 180U;
    const uint32_t bagVer190 = 190U;
    const uint32_t bagVer210 = 210U;
    const uint32_t bagVer230 = 230U;
    std::shared_ptr<mdc::ddt::Compressor> processor_;
    std::shared_ptr<rtftools::utils::NetProcess> netProcessPtr_;
};
}  // namespace rtfbag
}  // namespace rtf
#endif  // RTF_BAG_FILE_H
