/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 * Description: Define types in rtftools
 * Create: 2020-10-24
 */
#ifndef RTF_TOOLS_TYPES_H
#define RTF_TOOLS_TYPES_H

#include "ara/core/map.h"
#include "ara/core/string.h"
#include "ara/core/vector.h"

namespace rtf {
constexpr std::uint16_t RTF_DEFAULT_DOMAINID {1000U};
namespace rtfevent {
struct ResourceInfo {
    ara::core::String pubResourceName;
    ara::core::String subResourceName;
};

/**
 * @brief the class to store basic event info
 * @example Name Example <br>
 * EventName = ${EventType}[${InstanceShortname}] <br>
 * @note instanceShortName_ could be empty
 */
class RtfEventInfo {
public:
    RtfEventInfo() = default;
    ara::core::Vector<ara::core::String> GetSubs() const
    {
        return subList_;
    }
    void SetSubs(ara::core::Vector<ara::core::String> subs)
    {
        subList_.swap(subs);
    }
    ara::core::String GetPub() const
    {
        return pub_;
    }
    void SetPub(const ara::core::String pub)
    {
        pub_ = pub;
    }
    ara::core::String GetEventName() const
    {
        auto res = eventType_;
        if (!instanceShortName_.empty()) {
            res += '[' + instanceShortName_ + ']';
        }
        return res;
    }
    void SetEventName(const ara::core::String eventName)
    {
        ara::core::String eventNameTemp {eventName};
        ara::core::String instanceShortNameTemp {""};
        const auto leftBracket = eventName.find_first_of('[');
        const auto rightBracket = eventName.find_last_of(']');
        if (((leftBracket != ara::core::String::npos) && (rightBracket != ara::core::String::npos)) &&
            (leftBracket < rightBracket)) {
                eventNameTemp = eventName.substr(0U, leftBracket);
                instanceShortNameTemp = eventName.substr(leftBracket + 1UL, (rightBracket - leftBracket) - 1UL);
        }
        eventType_ = eventNameTemp;
        instanceShortName_ = instanceShortNameTemp;
    }
    const ara::core::String& GetInstanceShortName() const
    {
        return instanceShortName_;
    }
    void SetInstanceShortName(ara::core::String instanceName)
    {
        instanceShortName_ = std::move(instanceName);
    }
    ara::core::String GetEventType() const
    {
        return eventType_;
    }
    void SetEventType(const ara::core::String type)
    {
        eventType_ = type;
    }
    void SetAttribute(const ara::core::Map<ara::core::String, ara::core::String>& attributeList)
    {
        attributeList_ = attributeList;
    }
    ara::core::Map<ara::core::String, ara::core::String> GetAttribute() const
    {
        return attributeList_;
    }

    ara::core::String GetInstanceId() const { return instanceId_; }

    void SetInstanceId(const ara::core::String &instanceId) { instanceId_ = instanceId; }

    ara::core::String GetUuid() const { return uuid_; }

    void SetUuid(const ara::core::String &uuid) { uuid_ = uuid; }

    ara::core::String GetNetworkBinding() const { return networkBinding_; }

    void SetNetworkBinding(const ara::core::String &networkBinding) { networkBinding_ = networkBinding; }

    ara::core::String GetNetworkIP() const { return networkIP_; }

    void SetNetworkIP(const ara::core::String &networkIP) { networkIP_ = networkIP; }

    ara::core::String GetDomainId() const { return domainId_; }

    void SetDomainId(const ara::core::String &domainId) { domainId_ = domainId; }

    ara::core::String GetTransportPlugin() const { return transportPlugin_; }

    void SetTransportPlugin(const ara::core::String &transportPlugin) { transportPlugin_ = transportPlugin; }

    std::uint16_t GetServiceId() const noexcept { return serviceId_; };

    void SetServiceId(const uint16_t serviceId) { serviceId_ = serviceId; };

    ara::core::String GetTopicName() const noexcept { return topicName_;};

    void SetTopicName(const ara::core::String& topicName) { topicName_ = topicName; };

    ResourceInfo GetIAMResourceInfo() const noexcept { return resourceInfo_; }

    void SetIAMResourceInfo(const ResourceInfo &resourceInfo) { resourceInfo_ = resourceInfo; }

    ara::core::String GetUdpMulticastAddr() const { return udpMulticastAddr_; }

    void SetUdpMulticastAddr(const ara::core::String &udpMulticastAddr) { udpMulticastAddr_ = udpMulticastAddr; }

    ara::core::Vector<ara::core::String> GetDiscoveryMulticastAddr() const { return discoveryMulticastAddr_; }

    void SetDiscoveryMulticastAddr(const ara::core::Vector<ara::core::String> &discoveryMulticastAddr)
    {
        discoveryMulticastAddr_ = discoveryMulticastAddr;
    }

    ara::core::String GetFSHMMulticastAddr() const { return fshmMulticastAddr_; }

    void SetFSHMMulticastAddr(const ara::core::String &fshmMulticastAddr) { fshmMulticastAddr_ = fshmMulticastAddr; }

    bool GetFSHMMulticastStatus() const { return isFSHMMulticast_; }

    void SetFSHMMulticastStatus(const bool isFSHMMulticast) { isFSHMMulticast_ = isFSHMMulticast; }

private:
    ara::core::String pub_;
    ara::core::Vector<ara::core::String> subList_;
    ara::core::String eventType_;
    /** instance name is surrounded by '[' ']' */
    ara::core::String instanceShortName_;
    ara::core::Map<ara::core::String, ara::core::String> attributeList_;
    ara::core::String instanceId_;
    ara::core::String uuid_;
    ara::core::String networkBinding_;
    ara::core::String networkIP_;
    ara::core::String domainId_;
    ara::core::String transportPlugin_;
    ResourceInfo resourceInfo_;
    ara::core::String topicName_;
    std::uint16_t serviceId_{0U};
    ara::core::String udpMulticastAddr_;
    ara::core::Vector<ara::core::String> discoveryMulticastAddr_;
    bool isFSHMMulticast_{false};
    ara::core::String fshmMulticastAddr_;
};

class EventFilter {
public:
    enum class Type: uint8_t {
        PUBLISHER,
        SUBSCRIBER,
        ALL
    };
    enum class CommunicableType: uint8_t {
        UNDEFINED,
        LOCAL,
        NETWORK,
        CROSS,
        ALL
    };
    explicit EventFilter(const Type &type)
        : communicableType_(CommunicableType::UNDEFINED)
    {
        filterType_ = type;
    }
    void SetType(const Type &type)
    {
        filterType_ = type;
    }
    Type GetType() const
    {
        return filterType_;
    }
    void SetCommunicableType(const CommunicableType& type)
    {
        communicableType_ = type;
    }
    CommunicableType GetCommunicableType() const
    {
        return communicableType_;
    }
private:
    Type filterType_;
    CommunicableType communicableType_;
};
}

struct DataTypeReturnValue {
    ara::core::String name;
    ara::core::String type;
    ara::core::String dataType;
    ara::core::String value;
    ara::core::String isConstant;
    ara::core::String isOptional;
};

struct ShowReturnValue {
    ara::core::String dataTypeName;
    ara::core::String dataType;
};

class CommonFilter {
public:
    CommonFilter()
      : uuid_(""),
        networkIP_(""),
        domainId_(RTF_DEFAULT_DOMAINID)
    {
    }
    CommonFilter(const ara::core::String& uuid, const std::uint16_t& domainId)
    {
        uuid_ = uuid;
        domainId_ = domainId;
    }
    CommonFilter(const ara::core::String& uuid, const ara::core::String& networkIP, const std::uint16_t& domainId)
        : networkIP_(networkIP)
    {
        uuid_ = uuid;
        domainId_ = domainId;
    }
    ~CommonFilter() = default;
    void SetUuid(const ara::core::String& uuid)
    {
        uuid_ = uuid;
    }
    ara::core::String GetUuid() const
    {
        return uuid_;
    }
    void SetNetworkIP(const ara::core::String& networkIP)
    {
        networkIP_ = networkIP;
    }
    ara::core::String GetNetworkIP() const
    {
        return networkIP_;
    }
    void SetDomainId(const std::uint16_t& domainId)
    {
        domainId_ = domainId;
    }
    std::uint16_t GetDomainId() const
    {
        return domainId_;
    }
private:
    ara::core::String uuid_{""};
    ara::core::String networkIP_;
    std::uint16_t domainId_{RTF_DEFAULT_DOMAINID};
};
}
#endif
