/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 * Description:
 *      This file is the header of class RtfBagExtractor.
 *      RtfBagExtractor will read a bag file and extract specific event(s) from a bagfile
 * Create: 2020-07-16
 * Notes: NA
 */
#ifndef RTF_BAG_EXTRACTOR_H
#define RTF_BAG_EXTRACTOR_H

#include "rtf/internal/RtfBagFile.h"
#include "rtf/internal/RtfView.h"
#include "rtf/internal/RtfMsgEntity.h"

namespace rtf {
namespace rtfbag {
struct ExtractorOptions {
    ara::core::String inputFile{""};
    ara::core::String outputFile{""};
    ara::core::Vector<ara::core::String> events{};
    uint64_t startTime{0U};
    uint64_t endTime{0U};
    uint16_t domainId{RTF_DEFAULT_DOMAINID};
    ara::core::String networkIP;
};
class RtfBagExtractor {
public:
    enum class ExtractResult : uint8_t {
        RET_SUCCEED,
        RET_CONFLICT_FILE,
        RET_START_TIME_EXCCEED_LIMIT,
        RET_END_TIME_EXCCEED_LIMIT,
        RET_INVALID_TIME_RANGE,
        RET_NO_ENOUGH_SPACE,
        RET_NO_MESSAGE_FOUND,
        RET_READ_FAILED,
        RET_WRITE_FAILED,
        RET_FAILED,
        RET_INVALID_INPUT_FILE_NAME,
        RET_INVALID_OUTPUT_FILE_NAME,
        COMPRESS_ERROR,
        DECOMPRESS_ERROR,
        COMPRESS_CHUNK_ERROR,
        DECOMPRESS_CHUNK_ERROR,
        RENAME_TO_BAG_ERROR,
        RET_INVALID_DOMAIN_ID,
        RET_INVALID_NETWORK
    };

    enum class ExtractStatus : uint8_t {
            RET_IDLE,             // IDLE state, extract has not started or has ended
            RET_PREPARING,        // PREPAR state, preparation phase
            RET_EXTRACTING,       // EXTRACT state, split phase
            RET_STOPPING          // STOPPING state, stopping extract
    };

    struct ExtractProgress {
            ExtractStatus status;    // Current state
            double progress;         // Extract progress, [0, 1]
    };

    RtfBagExtractor(const ara::core::String& inputFile, const ara::core::String& outputFile,
                    const ara::core::Vector<ara::core::String>& events,
                    const uint64_t startTime, const uint64_t endTime);
    explicit RtfBagExtractor(const ExtractorOptions& options);
    ~RtfBagExtractor(void);

    RtfBagExtractor(const RtfBagExtractor &) = delete;
    RtfBagExtractor(RtfBagExtractor &&) = delete;
    RtfBagExtractor &operator=(const RtfBagExtractor &) & = delete;
    RtfBagExtractor &operator=(RtfBagExtractor &&) & = delete;

    ExtractResult Extract();
    ExtractProgress GetExtractProgress();
    void Stop() noexcept;
private:

    ExtractResult CheckParam(void) const;
    RtfView QueryViewFromBag(RtfBagFile& bagFile) const noexcept;
    ExtractResult WriteViewToFile(RtfView& view, const std::string& fileName) noexcept;
    ExtractResult WriteMsgEntityToBag(const RtfMsgEntity& msgEntity, RtfBagFile& bagFile) const noexcept;
    const ara::core::String UpdateOutputName(const ara::core::String& outputFile) const;
    void InitProgress();
    ExtractResult DoExtract();

    ExtractorOptions options_;
    ExtractProgress extractProgress_;
    std::mutex mutex_;
    rtf::rtfbag::FileCompressionType compressionType_;
    uint32_t version_;
};
}
}
#endif
