/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2021-2021. All rights reserved.
 * Description: Get the specified bag and event message information
 * Create: 2021-09-03
 */
#ifndef RTF_BAG_READER_H
#define RTF_BAG_READER_H

#include <map>
#include <memory>
#include <mutex>
#include <string>

#include "rtf/internal/RtfBagFile.h"
#include "rtf/internal/RtfLog.h"
#include "rtf/internal/RtfView.h"
#include "rtf/internal/rtf_tools_type.h"

namespace rtf {
namespace rtfbag {
class RtfBagReader {
public:
    enum class EventProtocol : std::uint8_t {
        DDS = 0x00U,
        SOMEIP = 0x01U,
        UNKNOWN = 0xFFU
    };
    enum class CompressionType : std::uint32_t {
        NONE = 0x00U,
        GZIP = 0x01U,
        ZLIB = 0x02U
    };
    enum class ErrorCode : std::uint8_t {
        SUCCESS = 0x00U, // 表示接口调用成功
        INIT_FAIL = 0x01U, // 表示初始化失败
        HAS_NEXT_FRAME = 0x02U, // 表示数据读取成功，并且读取列表中存在下一帧数据
        LAST_FRAME = 0x03U, // 表示数据读取成功，并且当前读取的数据帧为最后一帧数据
        WRONG_READ = 0x04U, // 表示读取到最后一帧数据后调用接口进行获取或者读取的指定长度超出消息的最大长度
        WRONG_INPUT = 0x05U, // 表示输入的Event或者指定的数据帧不存在
        DECOMPRESS_ERROR = 0x06U, // 表示解压存在失败
        MULTIEVENT_ERROR = 0x07U, // 表示存在同名event，无法读取指定内容
        UNKNOWN_ERROR = 0xFFU // 其他未定义错误
    };
    class RtfEventInfo {
    public:
        void SetDataType(const std::string &dataType) { dataType_ = dataType; }
        std::string GetDataType() const { return dataType_; }
        void SetProtocol(const EventProtocol protocol) { protocol_ = protocol; }
        EventProtocol GetProtocol() const { return protocol_; }
        void SetMsgCount(const std::uint32_t msgCount) { msgCount_ = msgCount; }
        std::uint32_t GetMsgCount() const { return msgCount_; }
        void SetDomainId(const std::uint16_t domainId) { domainId_ = domainId; }
        std::uint16_t GetDomainId() const { return domainId_; }
        void SetNetworkIP(const std::string& networkIP) { networkIP_ = networkIP; }
        std::string GetNetworkIP() const { return networkIP_; }
    private:
        std::string dataType_;
        EventProtocol protocol_ = EventProtocol::UNKNOWN;
        std::uint32_t msgCount_ = 0U;
        std::uint16_t domainId_ = 1000U;
        std::string networkIP_ ;
    };
    class RtfBagInfo {
    public:
        void SetBagVersion(const std::string &bagVersion) { bagVersion_ = bagVersion; }
        std::string GetBagVersion() const { return bagVersion_; }
        void AddToEventInfoList(const std::string &eventName,
            const RtfEventInfo &eventInfo) { eventInfoList_[eventName] = eventInfo; }
        const std::map<std::string, RtfEventInfo>& GetEventInfoList() const { return eventInfoList_; }

        void AddToMultiEventInfoList(const std::string &eventName,
            const RtfEventInfo &eventInfo) { static_cast<void>(multiEventInfoList_.emplace(std::pair<std::string, RtfEventInfo>(eventName, eventInfo))); }
        const std::multimap<std::string, RtfEventInfo>& GetMultiEventInfoList() const { return multiEventInfoList_; }

        void SetMsgCount(const std::uint32_t msgCount) { msgCount_ = msgCount; }
        std::uint32_t GetMsgCount() const { return msgCount_; }
        void SetBeginTime(const std::uint64_t beginTime) { beginTime_ = beginTime; }
        std::uint64_t GetBeginTime() const { return beginTime_; }
        void SetEndTime(const std::uint64_t endTime) { endTime_ = endTime; }
        std::uint64_t GetEndTime() const { return endTime_; }
        void SetBagSize(const std::uint64_t bagSize) { bagSize_ = bagSize; }
        std::uint64_t GetBagSize() const { return bagSize_; }
        void SetCompression(const CompressionType compression) { compression_ = compression; }
        CompressionType GetCompression() const { return compression_; }
    private:
        std::string bagVersion_ = "1.0";
        std::map<std::string, RtfEventInfo> eventInfoList_;
        std::multimap<std::string, RtfEventInfo> multiEventInfoList_;
        std::uint32_t msgCount_ = 0U;
        std::uint64_t beginTime_ = 0U;
        std::uint64_t endTime_ = 0U;
        std::uint64_t bagSize_ = 0U;
        CompressionType compression_ = CompressionType::NONE;
    };
    class EventMsg {
    public:
        void SetBuffer(std::vector<std::uint8_t> && buff) { buffer_ = std::move(buff); }
        const std::vector<std::uint8_t>& GetBuffer() const { return buffer_; }
        size_t GetBufferSize() const { return buffer_.size(); }
        void SetTimeStamp(const std::uint64_t timeStamp) { timeStamp_ = timeStamp; }
        std::uint64_t GetTimeStamp() const { return timeStamp_; }
        void SetEventName(const std::string &eventName) { eventName_ = eventName; }
        std::string GetEventName() const { return eventName_; }
    private:
        std::vector<std::uint8_t> buffer_;
        std::uint64_t timeStamp_ = 0U;
        std::string eventName_;
    };
     
    class RtfReaderIndex {
    public:
        RtfReaderIndex(const rtf::stdtype::String& name, rtf::stdtype::uint16_t domainId)
            : name_(name), domainId_(domainId) {}
        RtfReaderIndex(const rtf::stdtype::String& name, const rtf::stdtype::uint16_t domainId, const rtf::stdtype::String& networkIP)
            : name_(name), domainId_(domainId), networkIP_(networkIP) {}
        RtfReaderIndex() = delete;
        friend bool operator<(const RtfReaderIndex& lhs, const RtfReaderIndex& rhs) noexcept
        {
            if (lhs.name_ < rhs.name_) {
                return true;
            } else if (lhs.name_ == rhs.name_) {
                if (lhs.domainId_ < rhs.domainId_) {
                    return true;
                } else if (lhs.domainId_ == rhs.domainId_) {
                    if (lhs.networkIP_ < rhs.networkIP_) {
                        return true;
                    }
                } else {
                    // do nothing
                }
            }
            return false;
        }
        friend bool operator ==(const RtfReaderIndex& lhs, const RtfReaderIndex& rhs) noexcept
        {
            return ((lhs.name_ == rhs.name_) && (lhs.domainId_ == rhs.domainId_) &&
                   (lhs.networkIP_ == rhs.networkIP_));
        }
        inline rtf::stdtype::String GetName() const noexcept { return name_; }
        inline rtf::stdtype::uint16_t GetDomainId() const noexcept { return domainId_; }
        inline rtf::stdtype::String GetNetworkIP() const noexcept { return networkIP_; }
    private:
        rtf::stdtype::String name_;
        rtf::stdtype::uint16_t domainId_{1000U};
        rtf::stdtype::String networkIP_;
    };
    explicit RtfBagReader(const std::string& bagPath);
    ~RtfBagReader() = default;
    RtfBagReader(const RtfBagReader& other);
    RtfBagReader& operator=(const RtfBagReader& other) &;
    ErrorCode Init();
    RtfBagInfo GetBagInfo() const;
    // 按文件结构顺序读取一帧event的数据,调用一次读取一帧
    ErrorCode ReadEventMsg(EventMsg &eventMsg, const std::uint32_t len = 0U);
    // 读取指定event name的第N帧数据
    ErrorCode ReadEventMsg(const std::string &eventName, const std::uint32_t seq, EventMsg &eventMsg,
        const std::uint32_t len = 0U, const std::unique_ptr<rtf::CommonFilter> filter = nullptr);
    // 按时间序获取指定event name的一帧数据，调用一次读取一帧
    ErrorCode ReadEventMsg(const std::string &eventName, EventMsg &eventMsg, 
        const std::uint32_t len = 0U, const std::unique_ptr<rtf::CommonFilter> filter = nullptr);
private:
    RtfBagReader& operator=(RtfBagReader && other) & = delete;
    RtfBagReader(RtfBagReader && other) = delete;
    class EventMsgView {
    public:
        EventMsgView(RtfBagFile& bagfile, uint64_t const startTime, uint64_t const endTime,
            const std::string &eventname = "", const std::uint16_t domainId = 1000U);
        ~EventMsgView() = default;
        EventMsgView(const EventMsgView& other);
        EventMsgView& operator=(const EventMsgView& other) &;
        EventMsgView& operator++();
        RtfMsgEntity GetPosValue(const uint32_t pos) const { return eventMsgView_.Begin().GetPosValue(pos); }
        bool IsEnd() const { return eventMsgViewInter_.IsEnd(); }
        RtfMsgEntity Value() { return eventMsgViewInter_.Value(); }
    private:
        EventMsgView& operator=(EventMsgView && other) & = delete;
        EventMsgView(EventMsgView && other) = delete;
        RtfView eventMsgView_;
        RtfView::Iterator eventMsgViewInter_;
    };
    void CopyCtorToMakeEventViews(const RtfBagReader& other);
    void GetBagInfoAndMakeViews(const BagFileHeadInfo &headInfo);
    void MakeEventViews();
    ErrorCode GetNextEventMsg(EventMsg &eventMsg, EventMsgView &eventMsgView, const uint32_t len) const;
    ErrorCode ChangeToEventMsg(EventMsg &eventMsg, const RtfMsgEntity& msgEntity, const uint32_t len) const;
    std::string bagPath_;
    std::shared_ptr<RtfBagFile> rtfbagFile_;
    std::shared_ptr<EventMsgView> allBagView_;
    std::map<RtfReaderIndex, std::shared_ptr<EventMsgView>> eventMsgViewList_;
    std::once_flag initFlag_;
    ErrorCode initResult_;
    std::shared_ptr<RtfLog::Log> logger_ ;
    RtfBagInfo bagInfo_;
    // behavior mutex, can not support publish use multi thread
    std::mutex lock_;
};
}
}
#endif
