/* Copyright (c) Huawei Technologies Co., Ltd. 2019-2019. All rights reserved.
 * Description:
 *      This file is the defination of class RtfPlayer.
 * Create: 2019-12-12
 * Notes: NA
 */
#ifndef RTF_PLAYER_H
#define RTF_PLAYER_H

#ifndef RTFBAG_PLAY_CFLOAT
#define RTFBAG_PLAY_CFLOAT
#include <cfloat>
#endif
#include <chrono>
#include <functional>
#include <memory>
#include <tuple>

#include "ara/core/map.h"
#include "ara/core/string.h"
#include "ara/core/vector.h"
#include "rtf/com/rtf_com.h"
#include "rtf/internal/RtfBagFile.h"
#include "rtf/internal/RtfMsgEntity.h"
#include "rtf/internal/RtfPubClock.h"
#include "rtf/internal/RtfTimeTranslator.h"
#include "rtf/internal/RtfView.h"
#include "rtf/internal/tools_common_client_manager.h"
#include "rtf/internal/raw_buffer_for_raw_data.h"

namespace rtf {
namespace rtfbag {
struct PlayerOptions {
    PlayerOptions() noexcept = default;

    // 浮点数默认值设置最好是2的倍数或者1/2的倍数，否则可能会导致判断错误
    const std::uint32_t MAX_QUEUE_SIZE = 1000U;
    const std::uint32_t DEFAULT_QUEUE_SIZE = 100U;
    const std::uint32_t MIN_QUEUE_SIZE = 0U;

    const double MAX_FREQUENCY = 1000.0;
    const double DEFAULT_FREQUENCY = 0.0;
    const double MIN_FREQUENCY = 0.0;

    const double MAX_RATE = 50.0;
    const double DEFAULT_RATE = 1.0;
    const double MIN_RATE = 0.0625;

    const double MAX_DELAYTIME = 20.0;
    const double DEFAULT_DELAYTIME = 0.2;
    const double MIN_DELAYTIME = 0.0;

    const double MAX_STARTTIME = 10000.0;
    const double DEFAULT_STARTTIME = 0.0;
    const double MIN_STARTTIME = 0.0;

    const double MAX_DURATION = 10000000.0;
    const double DEFAULT_DURATION = 0.0;
    const double MIN_DURATION = 0.0;

    const double MAX_SKIPEMPTY = DBL_MAX;
    const double DEFAULT_SKIPEMPTY = DBL_MAX;
    const double MIN_SKIPEMPTY = 0.0;

    const std::uint32_t MAX_CHANGEMODE = 3U;
    const std::uint32_t DEFAULT_CHANGEMODE = 0U;
    const std::uint32_t MIN_CHANGEMODE = 0U;

    const std::uint16_t MAX_DOMAINID = 63U;
    const std::uint16_t DEFAULT_DOMAINID = 1000U;

    bool quiet = false;                        // 默认false
    bool immediate = false;                    // 默认false
    bool pause = false;                        // 默认false
    std::uint32_t queueSize = MIN_QUEUE_SIZE;  // 默认100，取值范围[0,1000]
    double frequency = DEFAULT_FREQUENCY;      // 默认0.0，取值范围[0.0,1000.0]
    double rate = DEFAULT_RATE;                // 默认1.0，取值范围[0.0625,50.0]
    double delayTime = DEFAULT_DELAYTIME;      // 默认0.2s，取值范围[0.0,20.0]
    double startTime = DEFAULT_STARTTIME;      // 默认0.0，取值范围[0.0,10000.0]
    double duration = DEFAULT_DURATION;        // 默认0.0，取值范围[0.0,10000000.0]
    double skipEmpty = DEFAULT_SKIPEMPTY;      // 默认DBL_MAX，取值范围[0.0,DBL_MAX]
    bool loop = false;                         // 默认为false
    bool adjustClock = false;
    bool accessControl = false;
    std::uint32_t changeMode = 0U;  // 更换通道，[0,3]
    ara::core::String port;         // 输入可选port区间，闭区间[num1,num2]
    ara::core::String someipNet;    // 输入SOMEIP network名称
    ara::core::String ddsNet;       // 输入DDS network ip地址
    std::uint16_t domainId = DEFAULT_DOMAINID;    // DDS domainId，默认值1000表示无效，取值范围[0,63]
    ara::core::String networkIP;    // network ip of event, for filter

    ara::core::Vector<ara::core::String> events;
    ara::core::Vector<ara::core::String> pauseEvents;
    ara::core::Vector<ara::core::String> skipEvents;
    ara::core::Vector<ara::core::String> bagFiles;
    ara::core::Map<ara::core::String, ara::core::String> forcePlayEvents; // key: eventName, value: ip
};

class RtfPlayer {
using RtfBagLog = ara::godel::common::log::Log;

public:
    enum class PlayerPrintType : uint8_t {
        OPEN_FILE  = 0U,     // 打开文件
        OPEN_FAIL,          // 文件打开失败
        NO_MSG,             // 文件中无可播放的消息
        WAITING,            // 延时等待提示
        WAIT_DONE,          // 延时等待结束
        HIT_PAUSE_STEP,     // 提示暂停或单步方法
        PLAYING,            // 播放过程提示，反复刷新同一行
        PLAY_DONE,          // 播放结束
        PORT_EROR,          // 可用端口不够或者错误
        NO_PORT,            // 没有输入端口区间
        NO_SOMEIPNET,       // 没有输入SOME/IP网卡
        SOMEIP_PUB_EROR,    // SOMEIP pub创建失败
        DDS_PUB_EROR,       // DDS pub创建失败
        E2E_INFO,           // 显示使用了E2E的events
        PARAMETER_EROR,     // 输入参数存在错误
        EVENTS_EXIST,       // 显示已经存在的events
        NO_MAINTAIND,       // 环境上不存在maintaind
        DECOMPRESS_EROR,    // 解压初始化失败
        DECOMPRESS_CHUNK_EROR, // 解压chunk数据失败
        ADJUST_CLOCK_DONE,     // 设置系统时钟成功
        ADJUST_CLOCK_FAIL,     // 设置系统时钟失败
        MAX_TYPE
    };

    struct PlayerEchoInfo {
        PlayerEchoInfo() = default;
        ara::core::String fileName;
        ara::core::String portSec;
        bool isRunning{ true };
        bool isTracing{ false };
        double delayTime{ 0.0 };
        double bagTime{ 0.0 };
        std::atomic<double> duration{ 0.0 };
        std::atomic<double> totalTime{ 0.0 };
        std::set<std::string> e2eEventNameList;
        std::set<std::string> serviceExistedEventList;
    };

    struct PortUse {
        rtf::stdtype::uint16_t port;
        bool inUse;
    };

    using PrintCallBack = std::function<void(PlayerPrintType, const PlayerEchoInfo&)>;

    explicit RtfPlayer(const PlayerOptions& options);
    ~RtfPlayer();
    RtfPlayer(const RtfPlayer &) = delete;
    RtfPlayer(RtfPlayer &&) = delete;
    RtfPlayer &operator=(const RtfPlayer &) & = delete;
    RtfPlayer &operator=(RtfPlayer &&) & = delete;

    bool Publish();
    void Pause();
    void Step();
    void Trace();
    void Stop();
    void RegPrintCallback(const PrintCallBack& callback);
    double GetPlayProgress() const;

private:
    void ShowThePlayerOptions() const noexcept;
    bool OpenRead();
    std::tuple<std::uint64_t, std::uint64_t> CalculateTime() const noexcept;
    bool Run(const std::uint64_t initTime, const std::uint64_t finishTime);
    bool DoRun(std::unique_ptr<RtfView>&& viewPtr);
    void PreTimeHandle();
    void SetSysTime();
    bool IsUsingRawData(const Connection& connection) const noexcept;
    bool CreatePublisher(const std::string& eventName, const Connection& connection);
    void Advertise(ara::core::Vector<Connection const *> const& connections) noexcept;
    bool DoPublish(RtfMsgEntity const& msgEntity, RtfBuffer& buffer);
    bool PreparePubMsg(const RtfBuffer& buffer, const bool isUsingRawData) noexcept;
    bool ChangePubMsgBeforeBagV200(const RtfBuffer& buffer) noexcept;
    bool ChangePubMsgForNormalMbuf(const RtfBuffer& buffer) noexcept;
    bool ChangePubMsgFromBagV200(const RtfBuffer& buffer) noexcept;
    bool ChangePubMsgForMbufChain(const RtfBuffer& buffer) noexcept;
    bool PrepareMbufChainHead(const RtfBuffer& buffer, std::uint64_t& offSet) noexcept;
    bool ChangePubMsg(const RtfBuffer& buffer) noexcept;
    bool PubMsg(std::uint64_t const time, const bool isUsingRawData);
    bool IsSkipEvent() const noexcept;
    void PauseStart();
    void PauseEvent();
    bool IsPauseEvent() const noexcept;
    bool PubImmediate(const bool isUsingRawData);
    bool PubSkipEmpty(const bool isUsingRawData);
    bool PubFrequency(const bool isUsingRawData);
    void ProcessPause() noexcept;
    void PubStep(const bool isUsingRawData);
    void SetMsgPubTime(double const time) noexcept;
    void SetMsgPubEventEntity(const Connection& conn) noexcept;
    bool DoPubMsg(const bool isUsingRawData) noexcept;
    void DDSConfig(const ara::core::String& eventName, const ara::core::String& dataType,
                   const rtf::maintaind::DDSEventInfo& ddsInfo, const bool isUsingRawData,
                   const rtf::maintaind::SerializeConfig& serializeConfig) noexcept;
    void DDSConfigNetInfo(std::unique_ptr<rtf::com::config::DDSEventConfig> &dds,
                          const rtf::maintaind::DDSEventInfo &ddsInfo) noexcept;
    void DDSForcePlayConfig(const std::unique_ptr<rtf::com::config::DDSEventConfig>& ddsConfig,
        const ara::core::String &eventName, rtf::com::dds::ParticipantQos& participantQos) const noexcept;
    void SetDDSWriterQosInfo(std::unique_ptr<rtf::com::config::DDSEventConfig> &dds,
                             const rtf::maintaind::DDSEventInfo &ddsInfo,
                             const ara::core::String &eventName) const noexcept;
    void SOMEIPEventConfig(const ara::core::String& eventName,
                           const ara::core::String& dataType,
                           const rtf::maintaind::SomeipEventInfo& someipInfo,
                           const rtf::maintaind::SerializeConfig& serializeConfig) noexcept;
    void SOMEIPServiceConfig(const ara::core::String& eventName,
                             const ara::core::String& dataType,
                             const rtf::maintaind::SomeipEventInfo& someipInfo,
                             const rtf::maintaind::SerializeConfig& serializeConfig) noexcept;
    void PrintInfo(rtf::maintaind::DriverType driverType, const ara::core::String &eventName) const;
    std::tuple<ara::core::Vector<Connection const *>, ara::core::Vector<Connection const *> >
    DevideConnections(const ara::core::Vector<Connection const *> &connections) const noexcept;
    bool AdvertiseEvents(const ara::core::Vector<Connection const *> &connectionsDDS,
                         const ara::core::Vector<Connection const *> &connectionsSOMEIP);
    bool InitSOMEIPPubPort(ara::core::Vector<Connection const *> const& connections) noexcept;
    void PrintSomeipInfo(const ara::core::String& eventName,
                         const ara::core::String& dataType,
                         const rtf::maintaind::SomeipEventInfo& someipInfo) noexcept;
    void PrintDdsInfo(const ara::core::String &eventName, const ara::core::String &dataType,
                      const rtf::maintaind::DDSEventInfo &ddsInfo) const noexcept;
    void InitNodeHandle() const noexcept;
    void CheckPortStatusInService(rtf::stdtype::uint16_t serviceId,
                                  const ara::core::Map<rtf::stdtype::uint16_t, PortUse>& portInfoWithIid) noexcept;
    bool CheckPortStatus();
    bool CheckPublishers() const;
    bool IsPortInUse(const rtf::stdtype::uint16_t port) noexcept;
    bool GetInputPorts(const std::string& port);
    bool CheckThePlayerOptions() noexcept;
    bool CheckInputPortInfo(const std::string& port) const noexcept;
    bool CheckPortValue(const ara::core::String& portVal) const noexcept;
    bool GetInputPubPort();
    void UpdatePortMapInService(rtf::stdtype::uint16_t serviceId,
            const ara::core::Map<rtf::stdtype::uint16_t, PortUse>& portInfoWithIid) noexcept;
    void UpdatePortMap() noexcept;

    std::shared_ptr<rtf::com::config::E2EConfig> QueryDDSE2EConfigInfo(
        const ara::core::String& eventName, const rtf::maintaind::DDSEventInfo& ddsInfo) noexcept;
    std::shared_ptr<rtf::com::config::E2EConfig> QuerySomeipE2EConfigInfo(
        const ara::core::String& eventName, const rtf::maintaind::SomeipEventInfo& someipInfo) noexcept;
    rtf::maintaind::E2EConfigInfo
    QueryE2EConfigInfoForAll(const rtf::maintaind::QuerySubEventInfoIndex &index) const noexcept;
    std::tuple<bool, ::rtf::maintaind::E2EConfigInfo>
    QueryE2EConfigInfoForOne(rtftools::common::RTFMaintaindToolsServiceProxy &proxy,
                             const rtf::maintaind::QuerySubEventInfoIndex &index) const noexcept;
    bool CheckSOMEIPEvent(const Connection* const conn,
                          const rtf::maintaind::EventInfoWithPubSub& eventElement) noexcept;
    bool CheckDDSEvent(const Connection* const conn, const rtf::maintaind::EventInfoWithPubSub& eventElement) const;
    bool CheckEvent(const ara::core::Vector<Connection const* > & connections,
                    const ara::core::Vector<rtf::maintaind::EventInfoWithPubSub>& eventInfoWithPubSubList);
    bool CheckService(const ara::core::Vector<Connection const* > & connections);
    void WaitForFlush() noexcept;

    PlayerOptions options_;
    ara::core::Vector<std::shared_ptr<RtfBagFile>> bags_;
    uint64_t      startTime_;
    bool          isPaused_;
    bool          isPauseStart_;
    bool          isDoStep_;
    std::atomic<bool> isStop_;
    bool          isPauseEventEnable_;      // is pause event enable
    bool          isPauseEvents_;        // is has --pause-events parameters

    std::shared_ptr<rtf::com::NodeHandle> node_;
    using PublisherMap = ara::core::Map<PlayedEventIndex, rtf::com::Publisher<rtf::com::ShapeShifter>>;
    using RawDataPublisherMap = ara::core::Map<PlayedEventIndex, rtf::com::Publisher<rtf::rtfbag::RawBufferForRawData>>;
    using PublisherPortMap = ara::core::Map<rtf::stdtype::uint16_t, ara::core::Map<rtf::stdtype::uint16_t, PortUse>>;
    PublisherMap publishers_;
    RawDataPublisherMap rawDataPublisher_;
    PublisherPortMap pubPorts_;
    rtf::stdtype::uint16_t minPort_;
    rtf::stdtype::uint16_t maxPort_;
    std::set<rtf::stdtype::uint16_t> portSet_;
    std::queue<rtf::stdtype::uint16_t> availablePortQue_;
    rtf::stdtype::uint16_t currentPort_;
    std::uint32_t requiredPorts_;
    ara::core::String pubEventName_;
    PlayedEventIndex playedEventIndex_;
    rtf::rtfbag::RawBufferForRawData pubMsg_;
    bool ddsCreateFailed_;
    bool someipCreateFailed_;
    std::atomic<bool> isPausedStart_;

    MilliTimePoint  pausedTime_;
    MilliTimePoint  sysHorizon_;
    MilliTimePoint  frePubStartTime_;
    PrintCallBack   printCallback_;
    PlayerEchoInfo  printInfo_;
    rtf::rtfbag::RtfPubClock       pubClock_;
    rtf::rtfbag::RtfTimeTranslator timeTranlator_;
    std::atomic<double> duration_;
    std::atomic<double> totalTime_;
    std::shared_ptr<rtf::rtftools::common::ToolsCommonClientManager> toolsCommonClientManager_ = nullptr;
    std::set<std::string> e2eEventNameList_;
    std::set<std::string> serviceDuplicatedList_;
    uint64_t decomErrorTime_;
    uint32_t curBagVersion_;
    const uint32_t allToUdp{1U};
    const uint32_t mbufToUdp{3U};
    std::once_flag adjustClockDone_;
    std::once_flag adjustClockFail_;
    std::shared_ptr<rtftools::utils::NetProcess> netProcessPtr_;
};
}  // namespace rtfbag
}  // namespace rtf
#endif // RTF_PLAYER_H
