/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2019-2019. All rights reserved.
 * Description:
 *      This file is the implement of class RtfRecorder.
 *      rtf recorder will record event info.
 * Create: 2019-12-03
 * Notes: NA
 */
#ifndef RTF_RECORDER_H
#define RTF_RECORDER_H

#include <chrono>
#include <condition_variable>
#include <fstream>
#include <functional>
#include <iomanip>
#include <list>
#include <memory>
#include <mutex>
#include <queue>
#include <set>
#include <sstream>
#include <sys/statvfs.h>
#include <sys/vfs.h>
#include <thread>
#include <unistd.h>
#include <vector>

#include "ara/core/map.h"
#include "ara/core/string.h"
#include "ara/core/vector.h"
#include "rtf/com/rtf_com.h"
#include "rtf/internal/RtfBagFile.h"
#include "rtf/internal/tools_common_client_manager.h"
#include "rtf/internal/RtfInnerType.h"
#include "rtf/rtfevent/RtfEvent.h"

namespace rtf {
namespace rtfbag {
struct RecorderOptions {
    RecorderOptions() = default;
    enum class CompressionType : std::uint8_t {
        NONE = 0U,  // 未压缩
        GZIP = 1U,  // GZIP
        ZLIB = 2U,  // ZLIB
    };

    bool recordAll{ false };
    bool appendDate{ false };
    ara::core::String prefix;
    ara::core::String name;
    std::uint32_t bufferSize{ 268435456U };  // the unit of bufferSize is Byte
    std::uint32_t chunkSize{ 786432U };
    std::uint32_t limit{ 10000000U };
    std::uint32_t skipFrame{ 0U };                         // 跳帧个数
    CompressionType compression{ CompressionType::NONE };  // 压缩类型
    bool split{ false };                                   // 分隔的个数
    std::uint64_t maxSize{ 0U };                           // 最大文件大小
    std::uint32_t maxSplits{ 0U };                         // 最大分割个数
    std::uint32_t maxDuration{ 0U };
    ara::core::Vector<ara::core::String> events;
    std::uint64_t minSpace{ 1073741824U };
    ara::core::String path;
    ara::core::String someipNet;
    ara::core::String ddsNet;
    ara::core::String   uuid;
    uint16_t            domainId{ 1000U };                // DDS domainId，默认值1000表示无效，取值范围[0,63]
    ara::core::String networkIP;
};

enum class RecorderType : std::uint8_t {
    INFO_SUBSCRIBE_TO_EVENT   = 0U,    // 订阅消息
    INFO_RECORDER_TO_FILE     = 1U,    // 记录消息到bag file
    INFO_CLOSE_TO_FILE        = 2U,    // 关闭当前bag file
    WARN_LESS_SPACE           = 3U,    // 磁盘空间告警提示
    WARN_RECORD_SAME_FILE     = 4U,    // 录制相同的文件名称
    WARN_UNSUBSCRIBE_TO_EVENT = 5U,    // 没有订阅消息
    WARN_OPEN_INVALID_PATH    = 6U,    // 打开非法的文件路径
    WARN_NO_SOMEIP_NET        = 7U,    // 没有输入SOME/IP网卡network名称
    EROR_CREATE_SOMEIP        = 8U,    // SOMEIP sub创建失败
    EROR_CREATE_DDS           = 9U,    // DDS sub创建失败
    EROR_PARAMETER            = 10U,   // 输入参数存在错误
    WARN_BUFFER_FULL          = 11U,   // 录制缓冲区满丢包
    WARN_COMPRESS             = 12U,   // 压缩初始化失败
    WARN_COMPRESS_CHUNK       = 13U,   // 压缩chunk失败
    EROR_CLOSE_TO_FILE        = 14U,   // 关闭当前bag file失败
    RENAME_TO_BAG_ERROR       = 15U,   // 重命名bag文件失败
};
enum class SubscribeType : std::uint8_t {
    SUBSCRIBE_SUCCESS       = 0U,    // 订阅成功
    WAITFOR_UNSUBSCRIBE     = 1U,    // 等待订阅取消
    UNSUBSCRIBE_SUCCESS     = 2U,    // 取消订阅成功
};
struct OutgoingMessage {
    std::uint64_t                       time;
    ara::core::String                   recordingName;
    ara::core::String                   eventName;
    ara::core::String                   type;
    rtf::maintaind::DriverType          driverType;
    // buff for nomal msg or the msg in format[PrivateLength][PrivateValue][MbufData]
    std::vector<uint8_t>                buff;
};

struct OutgoingQueue {
    OutgoingQueue(const ara::core::String &name, const std::queue<OutgoingMessage> &que, const std::uint64_t tm);

    ara::core::String               fileName;
    std::queue<OutgoingMessage>     queue;
    uint64_t                        time;
};

class RtfRecorder {
public:
    using RecorderCallback = std::function<void(RecorderType, ara::core::String)>;
    explicit RtfRecorder(const RecorderOptions& options);
    ~RtfRecorder() = default;
    RtfRecorder(const RtfRecorder &) = delete;
    RtfRecorder(RtfRecorder &&) = delete;
    RtfRecorder &operator=(const RtfRecorder &) & = delete;
    RtfRecorder &operator=(RtfRecorder &&) & = delete;

    int32_t Run();
    void Stop();
    void RecorderRegisterCallback(const RecorderCallback &callback);
private:
    using SubResource = std::pair<rtf::com::Subscriber, std::unique_ptr<rtf::com::ThreadGroup>>;
    using QueryEventInfo = rtf::maintaind::proxy::methods::QueryEventInfo;
    struct DDSEventIndexInfo {
        rtf::stdtype::uint16_t serviceId;
        rtf::stdtype::uint16_t instanceId;
        rtf::stdtype::String networkSegment;
        ara::core::String topic;
        rtf::stdtype::uint16_t domainId;
    };

    struct SOMEIPEventIndexInfo {
        rtf::stdtype::uint16_t serviceId;
        rtf::stdtype::uint16_t instanceId;
        rtf::stdtype::String networkSegment;
        rtf::stdtype::uint16_t eventId;
    };

    bool IsRepeatedEvent(const rtf::maintaind::EventInfoWithPubSub& eventInfoWithPubSub,
                         const ara::core::String& event) const;
    void DealWithSubscribedEvent(const ara::core::String& event,
        const rtf::maintaind::EventInfoWithPubSub& eventInfoWithPubSubList,
        const bool ddsExist,
        const bool someipExist);
    bool CheckTheRecorderOptions() const noexcept;
    bool CheckStop();
    void DoRecord();
    void DoRecordHandle();
    void BagHandle();
    void DeleteBuffQueue() noexcept;
    void StartWriting();
    void StopWriting();
    void CheckNumSplits();
    void UpdateFileNames();
    bool CheckSize();
    bool CheckDisk();
    void DoSplit(std::uint64_t time, std::uint64_t maxDuration);
    bool CheckDuration(std::uint64_t time);
    void CheckLimit(const ara::core::String& event);
    bool CheckSkipFrame(const ara::core::String& event);
    OutgoingMessage PrepareOutOutgoingMbufMessage(const ara::core::String& event, const rtf::com::RecvMemory& msg);
    std::vector<std::uint8_t>
    PrepareOutgoingMbufChain(const ara::core::Vector<rtf::OutMbufData> &outMbufDataList,
                             const std::vector<std::uint8_t> &privData) const;
    OutgoingMessage PrepareOutOutgoingMessage(const ara::core::String& event, const rtf::com::RecvMemory& msg);
    void DoQueue(const ara::core::String& event, const OutgoingMessage& outMsg);
    void SubscribeEvent(const ara::core::String& event);
    bool ShouldSubscribe(const ara::core::String& event) const noexcept;
    void Subscribe();
    void DoSubscribe(const ara::core::String& event,
        const ara::core::Vector<rtf::maintaind::EventInfoWithPubSub>& eventInfoWithPubSubList);
    std::tuple<bool, bool, bool> RegistConfig(const ara::core::String& event,
        const ara::core::Vector<rtf::maintaind::EventInfoWithPubSub>& eventInfoWithPubSubList);
    void GetPubSub(const QueryEventInfo::Output& output,
        ara::core::Vector<rtf::maintaind::EventInfoWithPubSub>& eventInfoWithPubSubList) const;
    void HandleSub(const ara::core::String& event, SubResource&& sub);
    bool IsSubscribed(const ara::core::String& event) const noexcept;
    void UnSubscribeEvent(const ara::core::String& event);
    void UnSubscribeCheck();
    ara::core::String TimeToStr() const;
    void SubscribeCallback(const ara::core::String& event, const rtf::com::RecvMemory& msg);
    /**
     * @brief The callback of subscriber for recording the msg with mbuf
     *
     * @param[in] event    The event name using mbuf
     * @param[in] msg      The msg with mbuf will be recorded
     */
    void SubscribeMbufCallback(const ara::core::String& event, const rtf::com::RecvMemory& msg);
    void GetEventList();
    bool IsQueried(const ara::core::String& eventName) const noexcept;
    ara::core::Vector<rtf::rtfevent::RtfEventInfo> FilterPreSubscribeEvent(
        const ara::core::Vector<rtf::rtfevent::RtfEventInfo>& eventInfoList) const;
    std::uint64_t GetCurrentRealTime() const noexcept;
    std::uint64_t GetCurrentMonotonicTime() const noexcept;
    std::uint64_t ChangeMonotonicToRealTime(const std::uint64_t monotonicTime) const noexcept;
    bool RecordReady() const noexcept;
    void Start();
    void PrintInfo(rtf::maintaind::DriverType driverType, const ara::core::String& event);
    void UnSubscribeAll();
    void GetAllDataType();
    void GetDataTypeRefList();
    bool WriteEventMsg(const OutgoingMessage& out);
    bool InsertMsgDef(const ara::core::String& eventName, const ara::core::String& dataTypeName);
    void SetStartTime() const;
    void SetStopTime() const;
    void InitStartTime() noexcept;
    void PrintE2ECheckResult(const ara::core::String& event, const rtf::com::SampleInfo& sampleInfo) const;
    SubResource SubscribeDdsRawDataEvent(const ara::core::String& event, const bool isEnableE2E);
    SubResource SubscribeCommonEvent(const ara::core::String& event, const bool isEnableE2E);
    rtf::maintaind::EventInfo tempEventInfo_;
    ara::core::String event_;
    RecorderOptions options_;
    ara::core::Vector<ara::core::String> events_;
    ara::core::Map<ara::core::String, rtf::rtfevent::RtfEventInfo> eventsPreSubscribeMap_;
    std::set<ara::core::String> subCreateFailed_;
    std::set<ara::core::String> allEvents_;
    bool ddsCreateError_;
    bool someipCreateError_;
    ara::core::String pureFileName_;
    ara::core::String targetFileName_;
    ara::core::String writeFileName_;
    uint64_t splitCount_;
    uint64_t queueSize_;
    std::unique_ptr<std::queue<OutgoingMessage>> queue_;
    uint64_t startTime_;
    uint64_t startRecordRealTime_;
    uint64_t startRecordMonotonicTime_;
    uint64_t checkDiskNext_;
    uint64_t receiveNum_;
    bool writeEnabled_;
    uint64_t dropMsgNum_;
    uint64_t compressErrorTime_;
    std::mutex mutex_;
    std::mutex dataTypeMutex_;
    std::mutex stopMutex_;
    std::mutex eventsMutex_;
    std::mutex bagFileMutex_;
    std::mutex subscribeStatusMutex_;
    std::thread thread_;
    std::condition_variable_any queueCondition_;
    RecorderCallback recorderCallback_;     // 回调函数保存
    std::list<ara::core::String> currentFiles_;         // 当前文件保存
    std::shared_ptr<rtf::rtfbag::RtfBagFile> bag_;
    uint64_t lastBufferWarnTime_;
    std::set<ara::core::String> currentlyRecording_;
    std::shared_ptr<rtf::com::NodeHandle> nh_;
    std::atomic<bool> doRecordEnabled_;
    bool isBagCreated_;
    ara::core::Map<ara::core::String, uint32_t> limit_;
    ara::core::Map<ara::core::String, uint32_t> skipFrame_;
    ara::core::Map<ara::core::String, bool> recordErrorEvent_;
    ara::core::Map<ara::core::String, DDSEventIndexInfo> ddsSubs_;
    ara::core::Map<ara::core::String, SOMEIPEventIndexInfo> someipSubs_;
    ara::core::Vector<rtf::rtfevent::RtfEventInfo> eventInfoList_;
    ara::core::Map<ara::core::String, rtf::maintaind::EventInfo> eventInfos_;
    ara::core::Map<ara::core::String, rtf::maintaind::SerializeConfig> eventSerializeConfigs_;
    ara::core::Map<ara::core::String, ara::core::String> types_;
    ara::core::Map<ara::core::String, rtf::maintaind::DriverType> driverTypes_;
    ara::core::Map<ara::core::String, SubResource> subs_;
    ara::core::Map<ara::core::String, SubscribeType> subscribeStatus_;
    ara::core::Map<ara::core::String, bool> unsubscribeStatus_;
    // key: eventName对应的DataTypeName, Value: 对应msg文件的全量信息
    ara::core::Map<ara::core::String, ara::core::String> eventToMsgDef_;
    ara::core::Map<ara::core::String, ara::core::String> dataTypeToDef_;
    // key:eventName对应的DataTypeName, value:对应的所有数据类型描述信息
    ara::core::Map<ara::core::String, ara::core::Vector<ara::core::String>> dataTypeToRefDataType_;
    std::shared_ptr<rtf::rtftools::common::ToolsCommonClientManager> toolsCommonClientManager_ ;
    const uint32_t bufferSize_;
    std::shared_ptr<rtftools::utils::NetProcess> netProcessPtr_;
};
}  // namespace rtfbag
}  // namespace rtf
#endif // RTF_RECORDER_H
