/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 * Description: rtf event echo API header
 * Create: 2020-07-29
 * Notes: N/A
 */

#ifndef RTFTOOLS_RTFEVENT_ECHO_H
#define RTFTOOLS_RTFEVENT_ECHO_H

#include <memory>
#include <mutex>
#include <string>
#include <vector>
#include <utility>
#include <functional>
#include <set>

#include "ara/core/string.h"
#include "ara/core/map.h"
#include "rtf/com/rtf_com.h"
#include "rtf/internal/RtfInnerType.h"
#include "rtf/internal/tools_common_client_manager.h"
#include "rtf/maintaind/serialize/impl_type_serializationnode.h"

namespace rtf {
namespace rtftools{
namespace utils{
class NetProcess;
}
}
namespace rtfevent {
struct EchoOptions {
    EchoOptions();
    ara::core::String ddsNetwork;
    ara::core::String someipNetwork;
    ara::core::String networkIP;
    bool isUntypeset;
    bool isNoarr;
    ara::core::String uuid;
    std::uint16_t domainId;
};
class SampleInfo {
public:
    enum class DeserializationRetCode : uint8_t {
        OK = 0U,
        UNDEFINED_ERROR,
        MISSING_MSG_INFO
    };
    SampleInfo() : echoResultStr_(),
                   echoE2EResultStr_(),
                   hasDeserializationError_(false),
                   deserializationRetCode_(DeserializationRetCode::OK),
                   typeInfo_() {}
    ~SampleInfo() = default;
    SampleInfo(const SampleInfo& other) = delete;
    SampleInfo& operator=(const SampleInfo& other) & = delete;

    SampleInfo(SampleInfo && other) = default;
    SampleInfo& operator=(SampleInfo && other) & = default;
    ara::core::String& GetEchoResult() { return echoResultStr_; }
    void SetEchoResult(ara::core::String && echoResult)
    {
        echoResultStr_ = std::move(echoResult);
        hasDeserializationError_ = echoResultStr_.empty();
    }
    ara::core::String GetE2EResult() const {  return echoE2EResultStr_; }
    void SetE2EResult(ara::core::String && e2eResult) { echoE2EResultStr_ = std::move(e2eResult); }
    bool GetIsDeserializationError() const { return hasDeserializationError_; }
    DeserializationRetCode GetDeserializationRetCode() const { return deserializationRetCode_; }
    void SetDeserializationRetCode(DeserializationRetCode && retCode) { deserializationRetCode_ = std::move(retCode); }
    ara::core::String GetDeserializationRetInfo() const
    {
        return typeInfo_;
    }
    void SetDeserializationRetInfo(ara::core::String && typeInfo)
    {
        typeInfo_ = std::move(typeInfo);
    }
private:
    ara::core::String echoResultStr_;
    ara::core::String echoE2EResultStr_;
    bool hasDeserializationError_;
    DeserializationRetCode deserializationRetCode_;
    ara::core::String typeInfo_;
};

class RtfEventEcho {
public:
    enum class EchoResult : uint8_t {
        RET_SUCCEED,
        RET_INVALID_EVENT_NAME,
        RET_INVALID_DDS_NETWORK,
        RET_INVALID_SOMEIP_NETWORK,
        RET_INIT_FAILED,
        RET_QUERY_FAILED,
        RET_SUBSCRIBE_FAILED,
        RET_NOT_SUPPORT_DATA_TYPE,
        RET_INVALID_UUID,
        RET_INVALID_DOMAINID,
        RET_INVALID_NETWORKIP
    };

    using EchoCallback = std::function<void(const ara::core::String&)>;
    using EchoCallbackEx = std::function<void(SampleInfo &&)>;
    using DeserializationCallback = std::function<bool(const std::string&,
                                                        const std::vector<uint8_t>&,
                                                        std::string&)>;

    RtfEventEcho(void);
    explicit RtfEventEcho(const DeserializationCallback& deserializationCallback);
    ~RtfEventEcho(void);

    RtfEventEcho(const RtfEventEcho& other) = delete;
    RtfEventEcho& operator=(const RtfEventEcho& other) & = delete;
    RtfEventEcho(RtfEventEcho&& other) = delete;
    RtfEventEcho& operator=(RtfEventEcho&& other) & = delete;

    EchoResult Echo(const ara::core::String& eventName,
                    const EchoOptions& options,
                    const EchoCallback& callback) noexcept;
    EchoResult Echo(const ara::core::String& eventName, const EchoOptions& options,
        const EchoCallbackEx& callback) noexcept;
    void Stop(const ara::core::String& eventName) noexcept;

private:
    using MaintaindProxy  = rtf::maintaind::proxy::RTFMaintaindToolsServiceProxy;
    using EventInfo       = rtf::maintaind::EventRegisterInfo;
    using DDSEventInfo    = rtf::maintaind::DDSEventInfo;
    using SOMEIPEventInfo = rtf::maintaind::SomeipEventInfo;
    using ServiceList     = ara::com::ServiceHandleContainer<MaintaindProxy::HandleType>;
    using NodeHandle      = rtf::com::NodeHandle;
    using Subscriber      = rtf::com::Subscriber;
    using SerializeType   = rtf::maintaind::SerializationType;

    enum class Protocol : std::uint8_t {
        DDS,
        SOMEIP,
        UNDEFINE
    };

    struct EchoRecord {
        ara::core::String echoEventName;
        std::unique_ptr<Subscriber> subscriber {nullptr};
        EchoCallback callback {nullptr};
        EchoCallbackEx echoCallbackEx {nullptr};
        Protocol protocol {Protocol::UNDEFINE};
        bool isStop {false};
        bool isRawBuffer {false};
        bool isRecvMemory {false};
    };

    static constexpr uint8_t JSON_STRING {0U};
    static constexpr uint8_t ROS_STRING {1U};
    static constexpr uint8_t NO_ARR_JSON {2U};
    static constexpr uint8_t NO_ARR_ROS {3U};
    static constexpr uint32_t ECHO_BUFFER_SIZE {10U};
    std::once_flag isInitialized_;
    std::unique_ptr<NodeHandle> nodeHandle_;
    EchoRecord echoRecord_;
    std::mutex echoMutex_;
    DeserializationCallback deserializationCallback_;
    SerializeType serializeType_;
    uint8_t formatType_;
    bool isEnableE2E_;
    bool isS2SService_;
    rtf::maintaind::PduConfigInfo pduInfo_;
    EchoOptions options_;
    std::unique_ptr<rtf::rtfevent::SampleInfo> sampleInfo_;
    std::shared_ptr<rtf::rtftools::common::ToolsCommonClientManager> toolsCommonClientManager_;
    ara::core::Optional<rtf::maintaind::SerializationNode> serializeRootNode_;
    std::shared_ptr<rtftools::utils::NetProcess> netProcessPtr_;

    RtfEventEcho(const DeserializationCallback& deserializationCallback, const SerializeType serializeType);
    bool Initialize(void);
    bool InitMaintaindConnection(void);
    bool InitializeRtfComInterface(void) noexcept;
    std::pair<bool, EchoResult> InitAndCheckValid(const ara::core::String& eventName, const EchoOptions& options);
    EchoResult CheckParam(const ara::core::String& eventName, const EchoOptions& options) const noexcept;
    std::unique_ptr<EventInfo> QueryEventInfoFromMaintaind(const ara::core::String& eventName);
    bool CheckMethodEntity(const rtf::maintaind::EventInfoWithPubSub& info);
    bool QueryTypeDefinitionFromMaintaind(
        ara::core::Map<ara::core::String, ara::core::String>& dataTypeJsonList) noexcept;

    EchoResult RegisterEvent(const ara::core::String& eventName,
                        const EchoOptions& options, const EventInfo& eventInfo) noexcept;
    bool SubscribeEvent(const ara::core::String& eventName,
                        const EventInfo& eventInfo) noexcept;
    std::unique_ptr<rtf::com::Subscriber> DoSubscribeRecvMemory(const ara::core::String& eventName,
        const ara::core::String& dataType);
    void OnEventReceived(
        const ara::core::String& eventName, const ara::core::String& dataType,
        const rtf::com::RecvMemory& data, const ara::core::Vector<rtf::OutMbufData>& mbuf = {}) noexcept;
    ara::core::String HandleEventName(const ara::core::String& eventName) const noexcept;
    void SetE2ECheckResult(const ara::core::String& eventName, const rtf::com::SampleInfo& sampleInfo);
    void RtfDeserializationToGetStr(const ara::core::String& eventName, const ara::core::String& dataType,
        std::vector<std::uint8_t> &&data, const ara::core::Vector<rtf::OutMbufData>& mbuf) noexcept;
    void CustomDeserializationToGetStr(const ara::core::String& eventName, std::vector<std::uint8_t> &&data) noexcept;
    ara::core::Vector<rtf::OutMbufData> GetMbufData(const rtf::com::RecvMemory& data) const noexcept;
    bool CheckIsNoSupportRawData(const RtfEventEcho::EventInfo& eveInfo) const;
    std::pair<bool, std::vector<std::uint8_t>> TruncateMsg(const std::vector<std::uint8_t>& data) const noexcept;
    std::string uuid_;
};
}
}

#endif
