/* Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 * Description: rtf method API
 * Create: 2020-12-03
 * Notes: NA
 */

#ifndef RTFTOOLS_RTFMETHOD_CALL_H
#define RTFTOOLS_RTFMETHOD_CALL_H

#include <chrono>
#include <cmath>
#include <deque>
#include <functional>
#include <mutex>
#include <numeric>
#include <set>

#include "ara/core/map.h"
#include "ara/core/string.h"
#include "ara/core/vector.h"
#include "json_parser/document.h"
#include "json_parser/global.h"
#include "rtf/com/rtf_com.h"
#include "rtf/internal/rtf_tools_type.h"
#include "rtf/internal/tools_common_client_manager.h"
#include "rtf/maintaind/serialize/impl_type_serializationnode.h"

namespace rtf {
namespace rtftools{
namespace utils{
class NetProcess;
}
}
namespace rtfmethodcall {
class BitBuffer;
using Document = ara::godel::common::jsonParser::Document;
using JsonParseValue = ara::godel::common::jsonParser::JsonParseValue;
struct MethodCallInfo {
    ara::core::String e2eCheckResult;
    ara::core::String replyResponse;
};

struct CallOptions {
    CallOptions();
    ara::core::Vector<ara::core::String> reqArgsStringList;
    ara::core::String inputParamJsonPath;
    ara::core::String ddsNet;
    ara::core::String someipNet;
    bool isUntypeset {false};
    bool isNoarr {false};
    ara::core::String uuid {""};
    std::uint16_t domainId {RTF_DEFAULT_DOMAINID};
    ara::core::String networkIP {""};
};

enum class RtfmethodCallType : uint8_t {
    ALL_RIGHT                      = 0U,    // 订阅消息
    WARN_NO_SOMEIP_NET             = 1U,    // 没有输入SOME/IP网卡network名称
    ERR_SERVER_STATUS              = 2U,    // method server不存在或者存在多个相同server
    ERR_REQ_MSG                    = 3U,    // 缺少入参的msg文件或序列化错误
    ERR_REP_MSG                    = 4U,    // 缺少出参的msg文件或反序列化错误
    ERR_INPUT_VALUE_NUM            = 5U,    // 输入参数个数存在错误
    ERR_INPUT_VALUE_TYPE           = 6U,    // 输入参数值存在错误
    ERR_REQUEST_TIME_OUT           = 7U,    // method请求超时
    ERR_REPLY_VALUE                = 8U,    // method返回值错误，触发流控或者异常导致
    ERR_CREATE_SOMEIP              = 9U,    // SOMEIP sub创建失败
    ERR_CREATE_DDS                 = 10U,   // DDS sub创建失败
    GET_RESPONSE                   = 11U,   // 获取到返回值
    NO_RESPONSE                    = 12U,   // 没有返回值，fireandforget场景或者返回值注册信息不全
    ERR_INPUT_FILE_PATH            = 13U,   // 输入json文件不存在
    ERR_CHECK_E2E                  = 14U,   // E2E保护结果存在错误时，将使用该错误码
    ERR_EXCEED_MAX_LENGTH          = 15U    // 序列化的数据超出长度字段最大长度
};

class RtfMethodCall {
    using QueryMethodInfo = rtf::maintaind::proxy::methods::QueryMethodInfo;
    using MethodRegisterInfo = rtf::maintaind::MethodRegisterInfo;
    using MethodInfoWithPubSub = rtf::maintaind::MethodInfoWithPubSub;
public:
    using CallResultCallback = std::function<void(MethodCallInfo, RtfmethodCallType)>;
    RtfMethodCall();
    bool Init(const ara::core::String &method, CallOptions callOptions, CallResultCallback callback);
    int32_t Call();
    void Stop();
private:
    class MethodType {
    public:
        struct Request {
            rtf::com::ShapeShifter data;
            static bool IsPlane()
            {
                return false;
            }
            using IsDpRawDataTag = void;
            using IsEnumerableTag = void;
            template<typename F>
            void enumerate(F& fun)
            {
                fun(data);
            }
            template<typename F>
            void enumerate(F& fun) const
            {
                fun(data);
            }
        };
        struct Response {
            rtf::com::ShapeShifter data;
            static bool IsPlane()
            {
                return false;
            }
            using IsDpRawDataTag = void;
            using IsEnumerableTag = void;
            template<typename F>
            void enumerate(F& fun)
            {
                fun(data);
            }
            template<typename F>
            void enumerate(F& fun) const
            {
                fun(data);
            }
        };
        Request  req;
        Response res;
    };
    bool InitInputParams();
    bool SerializeInputParams(MethodType& data);
    void CreateMethod(const ara::core::String& method);
    void FilterMethodList(const ara::core::String &pubName,
                          ara::core::Vector<MethodInfoWithPubSub> &methodInfoWithPubSubList);
    void PrintDDSConfig(rtf::maintaind::MethodRegisterInfo& ddsInfo) const noexcept;
    void DDSConfig(const ara::core::String& method, rtf::maintaind::MethodRegisterInfo& ddsInfo);
    void DDSConfigNetInfo(std::unique_ptr<rtf::com::config::DDSMethodConfig> &dds,
                          const ::rtf::maintaind::DDSMethodInfo& ddsInfo) noexcept;
    void SetTransportModeConfig(std::unique_ptr<rtf::com::config::DDSMethodConfig> &dds,
                                const rtf::maintaind::DDSMethodInfo& ddsMethodInfo,
                                const std::set<rtf::com::TransportMode>& transportModes) const noexcept;
    void SOMEIPConfig(const ara::core::String& method, rtf::maintaind::MethodRegisterInfo& someipInfo);
    void SOMEIPConfigNetInfo(std::unique_ptr<rtf::com::config::SOMEIPServiceConfig> &someipService,
                             std::unique_ptr<rtf::com::config::SOMEIPMethodConfig> &someipMethod,
                             const ::rtf::maintaind::SomeipMethodInfo& someipInfo) noexcept;
    void PrintSOMEIPConfig(const rtf::com::config::SOMEIPMethodConfig& someipMethod,
        const rtf::maintaind::MethodRegisterInfo& someipInfo,
        const rtf::maintaind::SomeipMethodInfo& someipMethodInfo) const noexcept;
    ara::core::String GetInstanceName(const ara::core::String &methodName) const;
    void CreateMethodClient(const ara::core::String& method,
        const ara::core::Vector<rtf::maintaind::MethodInfoWithPubSub>& methodInfoWithPubSubList);
    void CheckMethodClient(const ara::core::String& method, bool ddsExist, bool someipExist);
    void GetPubSub(const QueryMethodInfo::Output &outPut,
        ara::core::Vector<rtf::maintaind::MethodInfoWithPubSub>& methodInfoWithPubSubList) const;
    bool QueryTypeDefinitionFromMaintaind(
        ara::core::Map<ara::core::String, ara::core::String>& dataTypeJsonList) const noexcept;
    bool QueryAllTypeFromMaintaind();
    void PrintUserInfo(const MethodCallInfo& info, const RtfmethodCallType type) const;
    int32_t DealWithCallResult(const rtf::com::MethodClientResult& callResult,
                           const std::vector<std::uint8_t>& response,
                           const std::size_t outputTypeSize);
    int32_t CheckIsValidAndDoCall();
    int32_t DoCall();
    RtfmethodCallType CheckAndSetE2EResult(const rtf::com::MethodClientResult& callResult, MethodCallInfo& info) const;
    std::string DoDeserialize(const rtf::com::MethodClientResult& callResult, const rtf::com::ErrorCode err,
        const std::vector<std::uint8_t>& response);
    bool SerializeCommandLineInput(ara::core::Vector<rtf::rtfmethodcall::BitBuffer> &bitBufferList,
        std::size_t& dataLength);
    bool SerializeJsonFileInput(ara::core::Vector<rtf::rtfmethodcall::BitBuffer> &bitBufferList,
        std::size_t& dataLength);
    bool JsonInputNodeSerialize(const ara::core::Optional<rtf::maintaind::SerializationNode>& optionalSerializeNode,
                                size_t& right, size_t& dataLength,
                                ara::core::Vector<rtf::rtfmethodcall::BitBuffer> &bitBufferList);

    bool LineInputNodeSerialize(const ara::core::Optional<rtf::maintaind::SerializationNode>& optionalSerializeNode,
                                size_t& right, size_t& dataLength,
                                ara::core::Vector<rtf::rtfmethodcall::BitBuffer> &bitBufferList);
    void QueryReqTopStructMemberInfo();
    ara::core::String method_;
    CallOptions callOption_;
    rtf::com::MethodClient methodClient_ {};
    bool ddsCreateError_ {false};
    bool someipCreateError_ {false};
    ara::core::Vector<Document> documentList_;
    ara::core::Vector<ara::core::String> requestArgsTypeString_;
    std::vector<bool> reqTopStructMemberIsOptional_; // key is type, value is Optional mask
    ara::core::String replyOutputName_;
    ara::core::String requestInputName_;
    ara::core::String fieldName_;
    ara::core::String domainName_;
    rtf::maintaind::DriverType protocol_ {rtf::maintaind::DriverType::DDS};
    rtf::maintaind::SerializationType serializeTypeRequest_ {rtf::maintaind::SerializationType::CM};
    rtf::maintaind::SerializationType serializeTypeReply_ {rtf::maintaind::SerializationType::CM};
    rtf::maintaind::StructSerializationPolicy structPolicyRequest_ {rtf::maintaind::StructSerializationPolicy::OUT_LAYER_ALIGNMENT};
    rtf::maintaind::StructSerializationPolicy structPolicyReply_ {rtf::maintaind::StructSerializationPolicy::OUT_LAYER_ALIGNMENT};
    rtf::maintaind::SerializeConfigRawData requestSerializeConfig_;
    rtf::maintaind::SerializeConfigRawData replySerializeConfig_;
    std::mutex mutex_;
    std::shared_ptr<rtf::com::NodeHandle> nh_;
    CallResultCallback callResultCallback_;     // 回调函数保存
    bool isInit_ = false;
    bool isFireAndForget_ = false;
    bool isEnableE2E_ = false;
    static constexpr uint8_t JSON_STRING = 0U;
    static constexpr uint8_t ROS_STRING = 1U;
    static constexpr uint8_t NO_ARR_JSON = 2U;
    static constexpr uint8_t NO_ARR_ROS = 3U;
    uint8_t formatType_ {JSON_STRING};
    RtfmethodCallType serializeCallType_ = RtfmethodCallType::ERR_REQ_MSG;
    const std::map<rtf::com::e2e::ProfileCheckStatus, std::string> profileStatusMappingToString_ = {
        {rtf::com::e2e::ProfileCheckStatus::kOk, "kOk"},
        {rtf::com::e2e::ProfileCheckStatus::kRepeated, "kRepeated"},
        {rtf::com::e2e::ProfileCheckStatus::kWrongSequence, "kWrongSequence"},
        {rtf::com::e2e::ProfileCheckStatus::kError, "kError"},
        {rtf::com::e2e::ProfileCheckStatus::kNotAvailable, "kNotAvailable"},
        {rtf::com::e2e::ProfileCheckStatus::kNoNewData, "kNoNewData"},
        {rtf::com::e2e::ProfileCheckStatus::kCheckDisabled, "kCheckDisabled"}
    };
    std::shared_ptr<rtf::rtftools::common::ToolsCommonClientManager> toolsCommonClientManager_ = nullptr;
    std::shared_ptr<rtftools::utils::NetProcess> netProcessPtr_;
};
}
}
#endif // RTFTOOLS_RTFMETHOD_CALL_H
