/* Copyright (c) Huawei Technologies Co., Ltd. 2020-2020. All rights reserved.
 * Description: rtf node API .h file
 * Create: 2020-02-07
 * Notes:
 */

#ifndef RTFTOOLS_RTFNODE_H
#define RTFTOOLS_RTFNODE_H

#include <chrono>
#include <cmath>
#include <functional>
#include <mutex>

#include "ara/core/string.h"
#include "ara/core/vector.h"
#include "rtf/internal/tools_common_client_manager.h"
#include "rtf/stdtype/impl_type_int32_t.h"

namespace rtf {
namespace rtfnode {
/**
 * the class to store basic node info
 * @example Name Example <br>
 * ApplicationName = ${NodeName}[${HostName}] <br>
 * @note HostName should never be empty, unless you input an error value
 */
class RtfNodeInfo {
public:
    class EntityInfo {
    public:
        EntityInfo(const ara::core::String& name, const ara::core::String& uuid) {
            name_ = name;
            uuid_ = uuid;
        }
        ara::core::String GetName() const { return name_; }
        void SetName(const ara::core::String& name) { name_ = name; }
        ara::core::String GetUuid() const { return uuid_; }
        void SetUuid(const ara::core::String& uuid) { uuid_ = uuid; }
    private:
        ara::core::String name_;
        ara::core::String uuid_;
    };
    RtfNodeInfo() = default;
    ara::core::Vector<ara::core::String> GetServicInstance() const;
    void SetServiceInstances(ara::core::Vector<ara::core::String> serviceInstances) noexcept;

    ara::core::Vector<ara::core::String> GetEventPubList() const;
    void SetPubs(ara::core::Vector<ara::core::String> pubs) noexcept;
    ara::core::Vector<EntityInfo> GetEventPubInfoList() const;
    void SetEventPubInfoList(ara::core::Vector<EntityInfo> pubs);

    ara::core::Vector<ara::core::String> GetEventSubList() const;
    void SetSubs(ara::core::Vector<ara::core::String> subs) noexcept;
    ara::core::Vector<EntityInfo> GetEventSubInfoList() const;
    void SetEventSubInfoList(ara::core::Vector<EntityInfo> subs);

    ara::core::Vector<ara::core::String> GetMethodList() const;
    void SetMethods(ara::core::Vector<ara::core::String> methods) noexcept;
    ara::core::Vector<EntityInfo> GetMethodInfoList() const;
    void SetMethodInfoList(ara::core::Vector<EntityInfo> methods);

    ara::core::Vector<ara::core::String> GetFieldList() const;
    void SetFields(ara::core::Vector<ara::core::String> fields) noexcept;
    ara::core::Vector<EntityInfo> GetFieldInfoList() const;
    void SetFieldInfoList(ara::core::Vector<EntityInfo> fields);

    int32_t GetPidValue() const noexcept;
    void SetPidValue(const int32_t value) noexcept;

    ara::core::String GetApplicationName() const;
    void SetApplicationName(const ara::core::String& applicationName);

    const ara::core::String& GetNodeName() const noexcept;
    void SetNodeName(const ara::core::String& nodeName);

    const ara::core::String& GetHostName() const noexcept;
    void SetHostName(const ara::core::String& hostName);

    const ara::core::String& GetPartitionName() const noexcept;
    void SetPartitionName(const ara::core::String& partitionName);

    const ara::core::String& GetIdentity() const noexcept;
    void SetIdentity(const ara::core::String& identity);

    const ara::core::String& GetSelinuxLabel() const noexcept;
    void SetSelinuxLabel(const ara::core::String& selinuxLabel);

private:
    ara::core::Vector<ara::core::String> serviceInstanceList_;
    ara::core::Vector<ara::core::String> eventPubList_;
    ara::core::Vector<ara::core::String> eventSubList_;
    ara::core::Vector<ara::core::String> methodList_;
    ara::core::Vector<EntityInfo> eventPubInfoList_;
    ara::core::Vector<EntityInfo> eventSubInfoList_;
    ara::core::Vector<EntityInfo> methodInfoList_;
    ara::core::Vector<EntityInfo> fieldInfoList_;
    ara::core::Vector<ara::core::String> fieldList_;
    int32_t pid_ = 0;
    ara::core::String nodeName_;
    ara::core::String hostName_;
    ara::core::String partitionName_;
    ara::core::String identity_;
    ara::core::String selinuxLabel_;
};

class RtfNode {
    using QueryNodeInfo = rtf::maintaind::proxy::methods::QueryNodeInfo;
public:
    RtfNode();
    int32_t Query(const ara::core::String nodeName, RtfNodeInfo &rtfNodeInfo);
    int32_t QueryAll(ara::core::Vector<RtfNodeInfo> &rtfNodeInfoList);
    int32_t QueryAllWithNamespace(const ara::core::String &nodeNamespace, ara::core::Vector<RtfNodeInfo> &rtfNodeInfoList);
    int32_t Init();

private:
    bool GetQueryInfoResult(const QueryNodeInfo::Output& outPut, RtfNodeInfo &rtfNodeInfo) const;
    void GetQueryListResult(const QueryNodeInfo::Output& outPut, QueryNodeInfo::Output &nodeInfoListTmp) const;
    void SetQueryListResultWithFilter(const QueryNodeInfo::Output& outPut,
                                      const ara::core::String& nodeNamespace,
                                      ara::core::Vector<RtfNodeInfo> &rtfNodeInfoList) const;
    void SetQueryListResult(const QueryNodeInfo::Output& outPut,
                            ara::core::Vector<RtfNodeInfo> &rtfNodeInfoList) const;
    ara::core::Vector<RtfNodeInfo::EntityInfo> GetEntityInfoList(const rtf::maintaind::VectorNodeEntityInfo& infoList) const;
    bool isInit_ = false;
    std::shared_ptr<rtf::rtftools::common::ToolsCommonClientManager> toolsCommonClientManager_ = nullptr;
};
}
}
#endif
