/* Copyright (c) Huawei Technologies Co., Ltd. 2023-2023. All rights reserved.
 * Description: rtf method API
 * Create: 2023-05-20
 * Notes: NA
 */

#ifndef RTFTOOLS_RTFSERVICE_H
#define RTFTOOLS_RTFSERVICE_H

#include "ara/core/string.h"
#include "ara/core/vector.h"
#include "json_parser/json_writer.h"
#include "rtf/internal/tools_common_client_manager.h"
#include "rtf/internal/rtf_tools_type.h"
#include "rtf/rtfnode/RtfNode.h"

namespace rtf {
namespace rtftools{
namespace utils{
class NetProcess;
}
}
namespace rtfservice {
enum class ErrorCode : int32_t {
    SUCCESS = 0,
    FAILURE,
    INVALID_FILE_NAME,
    INVALID_PATH_NAME,
    WARN_WRITE_SAME_JSON_FILE,
    WARN_WRITE_SAME_ACTIVE_FILE,
    ERROR_FILE_CREAT_FAILED,
    ERROR_WRITE_DATA_FAILED,
    ERROR_RENAME_TO_JSON,
    ERROR_QUERY_FAILED
};

/**
 * the class to store basic method info
 * @example Name Example <br>
 * ServiceName = ${ServiceType}[${instanceShortname}] <br>
 * ServerName is ${AppName} also is ${Node}
 * @note instanceShortname could be empty
 */
class RtfServiceInfo {
public:
    RtfServiceInfo() = default;

    const ara::core::String GetServiceName() const noexcept;
    void SetServiceName(const ara::core::String& serviceName) noexcept;

    const ara::core::String GetServiceType() const noexcept;
    void SetServiceType(const ara::core::String &serviceType) noexcept;

    const ara::core::Vector<ara::core::String> GetFindServiceNode() const noexcept;
    void SetFindServiceNode(const ara::core::Vector<ara::core::String>& findServiceNode) noexcept;

    const ara::core::String GetOfferServiceNode() const noexcept;
    void SetOfferServiceNode(const ara::core::String &offerServiceNode) noexcept;

    ara::core::String GetNetworkBinding() const noexcept;
    void SetNetworkBinding(const ara::core::String &networkBinding) noexcept;

    const ara::core::String GetInstanceShortName() const noexcept;
    void SetInstanceShortName(const ara::core::String& instanceShortName) noexcept;

    ara::core::String GetNetworkIP() const noexcept;
    void SetNetworkIP(const ara::core::String &networkIP) noexcept;

    ara::core::String GetDomainId() const noexcept;
    void SetDomainId(const ara::core::String &domainId) noexcept;

    ara::core::String GetTransportPlugin() const noexcept;
    void SetTransportPlugin(const ara::core::String &transportPlugin) noexcept;

    ara::core::String GetInstanceId() const noexcept;
    void SetInstanceId(const ara::core::String &instanceId) noexcept;

    std::uint16_t GetServiceId() const noexcept;
    void SetServiceId(const std::uint16_t &serviceId) noexcept;

    ara::core::String GetIAMResourceName() const noexcept;
    void SetIAMResourceName(const ara::core::String& resourceName) noexcept;

private:
    ara::core::String serviceType_;
    ara::core::String serviceName_;
    ara::core::String instanceShortName_;
    ara::core::String offerServiceNode_;
    ara::core::Vector<ara::core::String> findServiceNode_;
    ara::core::String networkBinding_;
    ara::core::String networkIP_;
    ara::core::String domainId_;
    ara::core::String transportPlugin_;
    ara::core::String instanceId_;
    std::uint16_t serviceId_{0U};
    ara::core::String resourceName_;
};

class RtfService {
    using QueryServiceInfo = rtf::maintaind::proxy::methods::QueryServiceInfo;
public:
    RtfService();
    int32_t Query(const ara::core::String& serviceName, ara::core::Vector<RtfServiceInfo> &rtfServiceInfoList,
                  const std::unique_ptr<rtf::CommonFilter> filter = nullptr);
    int32_t QueryAll(ara::core::Vector<RtfServiceInfo> &rtfServiceInfoList);
    int32_t QueryAllWithNamespace(const ara::core::String nameSpace, ara::core::Vector<RtfServiceInfo> &rtfServiceInfoList);
    ErrorCode Export(const ara::core::String& fileName, const ara::core::String& path = "");
    int32_t Init();
private:
    ErrorCode InitExport();
    ErrorCode CheckInputParam();
    ErrorCode Write();
    ErrorCode WriteJson(const ara::core::Vector<RtfServiceInfo>& rtfServiceInfoList,
                        const ara::core::Vector<rtf::rtfnode::RtfNodeInfo>& nodeList) const;
    ErrorCode WriteServiceInfo(ara::godel::common::jsonParser::JsonWriter& json,
                               const RtfServiceInfo& rtfServiceInfo,
                               const ara::core::Vector<rtf::rtfnode::RtfNodeInfo>& nodeList) const;
    void BagFileSync() const;
    ErrorCode Stop() const;

    void CompareServiceInfo(bool &flag,
                          const rtf::maintaind::ServiceRegisterInfo& serviceInfoA,
                          const rtf::maintaind::ServiceRegisterInfo& serviceInfoB) const;
    void MergePubServiceInfoSubs(bool &flag,
                               const rtf::maintaind::ServiceInfoWithPubSub& serviceInfo,
                               ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoList) const;
    void MergeOffPubServiceInfoSubs(bool &flag,
                                  const rtf::maintaind::ServiceInfoWithPubSub& serviceInfo,
                                  ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoList) const;
    void MergeNoneNetworkServiceInfoSubs(bool &flag,
                                    const rtf::maintaind::ServiceInfoWithPubSub& serviceInfo,
                                    ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoList) const;
    void MergeNoPubServiceInfoSubs(std::size_t left,
                                 ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoList) const;
    void MergeServiceInfo(ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoList) const;
    void MergeNoPubServiceInfo(ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoList) const;

    void FilterServiceList(const ara::core::String &pubName,
                           const ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoList,
                           const CommonFilter& filter,
                           ara::core::Vector<RtfServiceInfo> &rtfServiceInfoList) const;
    void GetInfoResult(const QueryServiceInfo::Output &outPut,
                       ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoWithPubSubListTmp) const;
    void GetListResult(const ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceRegisterInfoList,
                       ara::core::Vector<RtfServiceInfo> &rtfServiceInfoList) const;
    void FilterResult(const ara::core::String &filtNamespace,
        const ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub>& serviceRegisterInfoList,
        ara::core::Vector<RtfServiceInfo> &rtfServiceInfoList) const;
    void StringSplit(const std::string& str, std::vector<std::string>& result) const;
    bool RequestServiceData(
        const std::shared_ptr<rtf::rtftools::common::ToolsCommonClientManager> &clientManager,
        ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoWithPubSubList) const;
    void RemoveOfflineServices(ara::core::Vector<rtf::maintaind::ServiceInfoWithPubSub> &serviceInfoList) const;
    void ExportServiceDdsInfo(ara::godel::common::jsonParser::JsonWriter& json, const RtfServiceInfo& rtfServiceInfo) const;
    bool isInit_ = false;
    std::string fileName_{};
    std::string targetFileName_{};
    std::string writeFileName_{};
    std::string pureFileName_{};
    std::string filePath_{};
    std::shared_ptr<rtf::rtftools::common::ToolsCommonClientManager> toolsCommonClientManager_ = nullptr;
    std::mutex lock_;
    bool isFileCreated_{false};
    FILE* file_{nullptr};
    std::shared_ptr<rtftools::utils::NetProcess> netProcessPtr_;
};
}
}
#endif // RTFTOOLS_RTFSERVICE_H
